@@ -8,79 +8,104 @@
 # @author Davide Brunato <brunato@sissa.it>
 #
 import json
-from collections.abc import Iterator
+from io import IOBase, TextIOBase
+from typing import Any, Dict, List, Optional, Type, Union, Tuple, \
+    IO, BinaryIO, TextIO, Iterator
 
 from .exceptions import XMLSchemaTypeError, XMLSchemaValueError, XMLResourceError
 from .names import XSD_NAMESPACE, XSI_TYPE
 from .etree import ElementTree, etree_tostring
+from .aliases import ElementType, XMLSourceType, NamespacesType, LocationsType, \
+    LazyType, SchemaSourceType, ConverterType, DecodeType, EncodeType, \
+    JsonDecodeType
 from .helpers import is_etree_document
 from .resources import fetch_schema_locations, XMLResource
 from .validators import XMLSchema10, XMLSchemaBase, XMLSchemaValidationError
 
 
-def get_context(source, schema=None, cls=None, locations=None, base_url=None,
-                defuse='remote', timeout=300, lazy=False, dummy_schema=False):
+def get_context(xml_document: Union[XMLSourceType, XMLResource],
+                schema: Optional[Union[XMLSchemaBase, SchemaSourceType]] = None,
+                cls: Optional[Type[XMLSchemaBase]] = None,
+                locations: Optional[LocationsType] = None,
+                base_url: Optional[str] = None,
+                defuse: str = 'remote',
+                timeout: int = 300,
+                lazy: LazyType = False,
+                dummy_schema: bool = False) -> Tuple[XMLResource, XMLSchemaBase]:
     """
     Get the XML document validation/decode context.
 
     :return: an XMLResource instance and a schema instance.
     """
+    resource: XMLResource
+    kwargs: Dict[Any, Any]
+
     if cls is None:
         cls = XMLSchema10
-    if not isinstance(source, XMLResource):
-        source = XMLResource(source, base_url, defuse=defuse, timeout=timeout, lazy=lazy)
-    if isinstance(schema, XMLSchemaBase) and source.namespace in schema.maps.namespaces:
-        return source, schema
-    if isinstance(source, XmlDocument) and source.schema is not None:
-        return source, source.schema
+    elif not issubclass(cls, XMLSchemaBase):
+        raise XMLSchemaTypeError(f"invalid schema class {cls}")
+
+    if isinstance(xml_document, XMLResource):
+        resource = xml_document
+    else:
+        resource = XMLResource(xml_document, base_url, defuse=defuse,
+                               timeout=timeout, lazy=lazy)
+
+    if isinstance(schema, XMLSchemaBase) and resource.namespace in schema.maps.namespaces:
+        return resource, schema
+    if isinstance(resource, XmlDocument) and isinstance(resource.schema, XMLSchemaBase):
+        return resource, resource.schema
 
     try:
-        schema_location, locations = fetch_schema_locations(source, locations, base_url=base_url)
+        schema_location, locations = fetch_schema_locations(resource, locations, base_url=base_url)
     except ValueError:
         if schema is None:
-            if XSI_TYPE in source.root.attrib:
-                schema = cls.meta_schema
+            if XSI_TYPE in resource.root.attrib and cls.meta_schema is not None:
+                return resource, cls.meta_schema
             elif dummy_schema:
-                return source, get_dummy_schema(source, cls)
+                return resource, get_dummy_schema(resource, cls)
             else:
                 msg = "no schema can be retrieved for the provided XML data"
                 raise XMLSchemaValueError(msg) from None
 
-        elif not isinstance(schema, XMLSchemaBase):
-            schema = cls(schema, validation='strict', locations=locations,
-                         base_url=base_url, defuse=defuse, timeout=timeout)
+        elif isinstance(schema, XMLSchemaBase):
+            return resource, schema
+        else:
+            return resource, cls(schema, locations=locations, base_url=base_url,
+                                 defuse=defuse, timeout=timeout)
     else:
-        schema = cls(schema or schema_location, validation='strict', locations=locations,
-                     defuse=defuse, timeout=timeout)
-
-    return source, schema
+        kwargs = dict(locations=locations, defuse=defuse, timeout=timeout)
+        if schema is None or isinstance(schema, XMLSchemaBase):
+            return resource, cls(schema_location, **kwargs)
+        else:
+            return resource, cls(schema, **kwargs)
 
 
-def get_dummy_schema(xml_resource, schema_class):
-    tag = xml_resource.root.tag
+def get_dummy_schema(resource: XMLResource, cls: Type[XMLSchemaBase]) -> XMLSchemaBase:
+    tag = resource.root.tag
     if tag.startswith('{'):
         namespace, name = tag[1:].split('}')
     else:
         namespace, name = '', tag
 
     if namespace:
-        return schema_class(
+        return cls(
             '<xs:schema xmlns:xs="{0}" targetNamespace="{1}">\n'
             '    <xs:element name="{2}"/>\n'
             '</xs:schema>'.format(XSD_NAMESPACE, namespace, name)
         )
     else:
-        return schema_class(
+        return cls(
             '<xs:schema xmlns:xs="{0}">\n'
             '    <xs:element name="{1}"/>\n'
             '</xs:schema>'.format(XSD_NAMESPACE, name)
         )
 
 
-def get_lazy_json_encoder(errors):
+def get_lazy_json_encoder(errors: List[XMLSchemaValidationError]) -> Type[json.JSONEncoder]:
 
     class JSONLazyEncoder(json.JSONEncoder):
-        def default(self, obj):
+        def default(self, obj: Any) -> Any:
             if isinstance(obj, Iterator):
                 while True:
                     result = next(obj, None)
@@ -93,9 +118,18 @@ def get_lazy_json_encoder(errors):
     return JSONLazyEncoder
 
 
-def validate(xml_document, schema=None, cls=None, path=None, schema_path=None,
-             use_defaults=True, namespaces=None, locations=None, base_url=None,
-             defuse='remote', timeout=300, lazy=False):
+def validate(xml_document: Union[XMLSourceType, XMLResource],
+             schema: Optional[XMLSchemaBase] = None,
+             cls: Optional[Type[XMLSchemaBase]] = None,
+             path: Optional[str] = None,
+             schema_path: Optional[str] = None,
+             use_defaults: bool = True,
+             namespaces: Optional[NamespacesType] = None,
+             locations: Optional[LocationsType] = None,
+             base_url: Optional[str] = None,
+             defuse: str = 'remote',
+             timeout: int = 300,
+             lazy: LazyType = False) -> None:
     """
     Validates an XML document against a schema instance. This function builds an
     :class:`XMLSchema` object for validating the XML document. Raises an
@@ -127,15 +161,24 @@ def validate(xml_document, schema=None, cls=None, path=None, schema_path=None,
     :class:`XMLResource` instances.
     :param lazy: optional argument for construct the :class:`XMLResource` instance.
     """
-    source, schema = get_context(
+    source, _schema = get_context(
         xml_document, schema, cls, locations, base_url, defuse, timeout, lazy
     )
-    schema.validate(source, path, schema_path, use_defaults, namespaces)
+    _schema.validate(source, path, schema_path, use_defaults, namespaces)
 
 
-def is_valid(xml_document, schema=None, cls=None, path=None, schema_path=None,
-             use_defaults=True, namespaces=None, locations=None, base_url=None,
-             defuse='remote', timeout=300, lazy=False):
+def is_valid(xml_document: Union[XMLSourceType, XMLResource],
+             schema: Optional[XMLSchemaBase] = None,
+             cls: Optional[Type[XMLSchemaBase]] = None,
+             path: Optional[str] = None,
+             schema_path: Optional[str] = None,
+             use_defaults: bool = True,
+             namespaces: Optional[NamespacesType] = None,
+             locations: Optional[LocationsType] = None,
+             base_url: Optional[str] = None,
+             defuse: str = 'remote',
+             timeout: int = 300,
+             lazy: LazyType = False) -> bool:
     """
     Like :meth:`validate` except that do not raises an exception but returns ``True`` if
     the XML document is valid, ``False`` if it's invalid.
@@ -146,9 +189,18 @@ def is_valid(xml_document, schema=None, cls=None, path=None, schema_path=None,
     return schema.is_valid(source, path, schema_path, use_defaults, namespaces)
 
 
-def iter_errors(xml_document, schema=None, cls=None, path=None, schema_path=None,
-                use_defaults=True, namespaces=None, locations=None, base_url=None,
-                defuse='remote', timeout=300, lazy=False):
+def iter_errors(xml_document: Union[XMLSourceType, XMLResource],
+                schema: Optional[XMLSchemaBase] = None,
+                cls: Optional[Type[XMLSchemaBase]] = None,
+                path: Optional[str] = None,
+                schema_path: Optional[str] = None,
+                use_defaults: bool = True,
+                namespaces: Optional[NamespacesType] = None,
+                locations: Optional[LocationsType] = None,
+                base_url: Optional[str] = None,
+                defuse: str = 'remote',
+                timeout: int = 300,
+                lazy: LazyType = False) -> Iterator[XMLSchemaValidationError]:
     """
     Creates an iterator for the errors generated by the validation of an XML document.
     Takes the same arguments of the function :meth:`validate`.
@@ -159,13 +211,22 @@ def iter_errors(xml_document, schema=None, cls=None, path=None, schema_path=None
     return schema.iter_errors(source, path, schema_path, use_defaults, namespaces)
 
 
-def to_dict(xml_document, schema=None, cls=None, path=None, process_namespaces=True,
-            locations=None, base_url=None, defuse='remote', timeout=300, lazy=False, **kwargs):
+def iter_decode(xml_document: Union[XMLSourceType, XMLResource],
+                schema: Optional[XMLSchemaBase] = None,
+                cls: Optional[Type[XMLSchemaBase]] = None,
+                path: Optional[str] = None,
+                validation: str = 'strict',
+                process_namespaces: bool = True,
+                locations: Optional[LocationsType] = None,
+                base_url: Optional[str] = None,
+                defuse: str = 'remote',
+                timeout: int = 300,
+                lazy: LazyType = False,
+                **kwargs: Any) -> Iterator[Union[Any, XMLSchemaValidationError]]:
     """
-    Decodes an XML document to a Python's nested dictionary. The decoding is based
-    on an XML Schema class instance. For default the document is validated during
-    the decoding phase. Raises an :exc:`XMLSchemaValidationError` if the XML document
-    is not validated against the schema.
+    Creates an iterator for decoding an XML source to a data structure. For default
+    the document is validated during the decoding phase and if it's invalid then one
+    or more :exc:`XMLSchemaValidationError` instances are yielded before the decoded data.
 
     :param xml_document: can be an :class:`XMLResource` instance, a file-like object a path \
     to a file or an URI of a resource or an Element instance or an ElementTree instance or \
@@ -177,6 +238,8 @@ def to_dict(xml_document, schema=None, cls=None, path=None, process_namespaces=T
     :class:`XMLSchema10`).
     :param path: is an optional XPath expression that matches the elements of the XML \
     data that have to be decoded. If not provided the XML root element is used.
+    :param validation: defines the XSD validation mode to use for decode, can be \
+    'strict', 'lax' or 'skip'.
     :param process_namespaces: indicates whether to use namespace information in \
     the decoding process.
     :param locations: additional schema location hints, in case a schema instance \
@@ -190,21 +253,57 @@ def to_dict(xml_document, schema=None, cls=None, path=None, process_namespaces=T
     :param lazy: optional argument for construct the :class:`XMLResource` instance.
     :param kwargs: other optional arguments of :meth:`XMLSchema.iter_decode` \
     as keyword arguments.
-    :return: an object containing the decoded data. If ``validation='lax'`` keyword argument \
-    is provided the validation errors are collected and returned coupled in a tuple with the \
-    decoded data.
-    :raises: :exc:`XMLSchemaValidationError` if the object is not decodable by \
-    the XSD component, or also if it's invalid when ``validation='strict'`` is provided.
+    :raises: :exc:`XMLSchemaValidationError` if the XML document is invalid and \
+    ``validation='strict'`` is provided.
     """
-    source, schema = get_context(
+    source, _schema = get_context(
         xml_document, schema, cls, locations, base_url, defuse, timeout, lazy
     )
-    return schema.decode(source, path=path, process_namespaces=process_namespaces, **kwargs)
+    yield from _schema.iter_decode(source, path=path, validation=validation,
+                                   process_namespaces=process_namespaces, **kwargs)
 
 
-def to_json(xml_document, fp=None, schema=None, cls=None, path=None, converter=None,
-            process_namespaces=True, locations=None, base_url=None, defuse='remote',
-            timeout=300, lazy=False, json_options=None, **kwargs):
+def to_dict(xml_document: Union[XMLSourceType, XMLResource],
+            schema: Optional[XMLSchemaBase] = None,
+            cls: Optional[Type[XMLSchemaBase]] = None,
+            path: Optional[str] = None,
+            validation: str = 'strict',
+            process_namespaces: bool = True,
+            locations: Optional[LocationsType] = None,
+            base_url: Optional[str] = None,
+            defuse: str = 'remote',
+            timeout: int = 300,
+            lazy: LazyType = False, **kwargs: Any) -> DecodeType[Any]:
+    """
+    Decodes an XML document to a Python's nested dictionary. Takes the same arguments
+    of the function :meth:`iter_decode`, but *validation* mode defaults to 'strict'.
+
+    :return: an object containing the decoded data. If ``validation='lax'`` is provided \
+    validation errors are collected and returned in a tuple with the decoded data.
+    :raises: :exc:`XMLSchemaValidationError` if the XML document is invalid and \
+    ``validation='strict'`` is provided.
+    """
+    source, _schema = get_context(
+        xml_document, schema, cls, locations, base_url, defuse, timeout, lazy
+    )
+    return _schema.decode(source, path=path, validation=validation,
+                          process_namespaces=process_namespaces, **kwargs)
+
+
+def to_json(xml_document: Union[XMLSourceType, XMLResource],
+            fp: Optional[IO[str]] = None,
+            schema: Optional[XMLSchemaBase] = None,
+            cls: Optional[Type[XMLSchemaBase]] = None,
+            path: Optional[str] = None,
+            converter: Optional[ConverterType] = None,
+            process_namespaces: bool = True,
+            locations: Optional[LocationsType] = None,
+            base_url: Optional[str] = None,
+            defuse: str = 'remote',
+            timeout: int = 300,
+            lazy: LazyType = False,
+            json_options: Optional[Dict[str, Any]] = None,
+            **kwargs: Any) -> JsonDecodeType:
     """
     Serialize an XML document to JSON. For default the XML data is validated during
     the decoding phase. Raises an :exc:`XMLSchemaValidationError` if the XML document
@@ -243,7 +342,7 @@ def to_json(xml_document, fp=None, schema=None, cls=None, path=None, converter=N
     :raises: :exc:`XMLSchemaValidationError` if the object is not decodable by \
     the XSD component, or also if it's invalid when ``validation='strict'`` is provided.
     """
-    source, schema = get_context(
+    source, _schema = get_context(
         xml_document, schema, cls, locations, base_url, defuse, timeout, lazy
     )
     if json_options is None:
@@ -253,21 +352,20 @@ def to_json(xml_document, fp=None, schema=None, cls=None, path=None, converter=N
     kwargs['converter'] = converter
     kwargs['process_namespaces'] = process_namespaces
 
-    errors = []
+    errors: List[XMLSchemaValidationError] = []
 
     if path is None and source.is_lazy() and 'cls' not in json_options:
         json_options['cls'] = get_lazy_json_encoder(errors)
 
-    obj = schema.decode(source, path=path, **kwargs)
+    obj = _schema.decode(source, path=path, **kwargs)
     if isinstance(obj, tuple):
+        errors.extend(obj[1])
         if fp is not None:
             json.dump(obj[0], fp, **json_options)
-            obj[1].extend(errors)
-            return tuple(obj[1])
+            return tuple(errors)
         else:
             result = json.dumps(obj[0], **json_options)
-            obj[1].extend(errors)
-            return result, tuple(obj[1])
+            return result, tuple(errors)
     elif fp is not None:
         json.dump(obj, fp, **json_options)
         return None if not errors else tuple(errors)
@@ -276,7 +374,12 @@ def to_json(xml_document, fp=None, schema=None, cls=None, path=None, converter=N
         return result if not errors else (result, tuple(errors))
 
 
-def from_json(source, schema, path=None, converter=None, json_options=None, **kwargs):
+def from_json(source: Union[str, bytes, IO[str]],
+              schema: XMLSchemaBase,
+              path: Optional[str] = None,
+              converter: Optional[ConverterType] = None,
+              json_options: Optional[Dict[str, Any]] = None,
+              **kwargs: Any) -> EncodeType[ElementType]:
     """
     Deserialize JSON data to an XML Element.
 
@@ -289,7 +392,8 @@ def from_json(source, schema, path=None, converter=None, json_options=None, **kw
     :param converter: an :class:`XMLSchemaConverter` subclass or instance to use \
     for the encoding.
     :param json_options: a dictionary with options for the JSON deserializer.
-    :param kwargs: Keyword arguments containing options for converter and encoding.
+    :param kwargs: other optional arguments of :meth:`XMLSchema.iter_encode` and \
+    options for converter.
     :return: An element tree's Element instance. If ``validation='lax'`` keyword argument is \
     provided the validation errors are collected and returned coupled in a tuple with the \
     Element instance.
@@ -301,10 +405,10 @@ def from_json(source, schema, path=None, converter=None, json_options=None, **kw
     elif json_options is None:
         json_options = {}
 
-    if hasattr(source, 'read'):
-        obj = json.load(source, **json_options)
-    else:
+    if isinstance(source, (str, bytes)):
         obj = json.loads(source, **json_options)
+    else:
+        obj = json.load(source, **json_options)
 
     return schema.encode(obj, path=path, converter=converter, **kwargs)
 
@@ -332,23 +436,34 @@ class XmlDocument(XMLResource):
     :param timeout: the timeout for base :class:`xmlschema.XMLResource` initialization.
     :param lazy: the lazy mode for base :class:`xmlschema.XMLResource` initialization.
     """
-    schema = _fallback_schema = None
-    validation = 'skip'
-    namespaces = None
-    errors = ()
+    schema: Optional[XMLSchemaBase] = None
+    _fallback_schema: Optional[XMLSchemaBase] = None
+    validation: str = 'skip'
+    namespaces: Optional[NamespacesType] = None
+    errors: Union[Tuple[()], List[XMLSchemaValidationError]] = ()
 
-    def __init__(self, source, schema=None, cls=None, validation='strict',
-                 namespaces=None, locations=None, base_url=None, allow='all',
-                 defuse='remote', timeout=300, lazy=False):
+    def __init__(self, source: XMLSourceType,
+                 schema: Optional[Union[XMLSchemaBase, SchemaSourceType]] = None,
+                 cls: Optional[Type[XMLSchemaBase]] = None,
+                 validation: str = 'strict',
+                 namespaces: Optional[NamespacesType] = None,
+                 locations: Optional[LocationsType] = None,
+                 base_url: Optional[str] = None,
+                 allow: str = 'all',
+                 defuse: str = 'remote',
+                 timeout: int = 300,
+                 lazy: LazyType = False) -> None:
 
+        if cls is None:
+            cls = XMLSchema10
         self.validation = validation
         self._namespaces = namespaces
         super(XmlDocument, self).__init__(source, base_url, allow, defuse, timeout, lazy)
 
         if isinstance(schema, XMLSchemaBase) and self.namespace in schema.maps.namespaces:
             self.schema = schema
-        elif schema is not None:
-            self.schema = (cls or XMLSchema10)(
+        elif schema is not None and not isinstance(schema, XMLSchemaBase):
+            self.schema = cls(
                 source=schema,
                 base_url=base_url,
                 allow=allow,
@@ -356,9 +471,6 @@ class XmlDocument(XMLResource):
                 timeout=timeout,
             )
         else:
-            if cls is None:
-                cls = XMLSchema10
-
             try:
                 schema_location, locations = fetch_schema_locations(self, locations, base_url)
             except ValueError:
@@ -379,14 +491,16 @@ class XmlDocument(XMLResource):
                     timeout=timeout,
                 )
 
-        if validation == 'strict':
+        if self.schema is None:
+            pass
+        elif validation == 'strict':
             self.schema.validate(self, namespaces=self.namespaces)
         elif validation == 'lax':
             self.errors = [e for e in self.schema.iter_errors(self, namespaces=self.namespaces)]
         elif validation != 'skip':
             raise XMLSchemaValueError("{!r}: not a validation mode".format(validation))
 
-    def parse(self, source, lazy=False):
+    def parse(self, source: XMLSourceType, lazy: LazyType = False) -> None:
         super(XmlDocument, self).parse(source, lazy)
         self.namespaces = self.get_namespaces(self._namespaces)
 
@@ -397,11 +511,11 @@ class XmlDocument(XMLResource):
         elif self.validation == 'lax':
             self.errors = [e for e in self.schema.iter_errors(self, namespaces=self.namespaces)]
 
-    def getroot(self):
+    def getroot(self) -> ElementType:
         """Get the root element of the XML document."""
         return self._root
 
-    def get_etree_document(self):
+    def get_etree_document(self) -> Any:
         """
         The resource as ElementTree XML document. If the resource is lazy raises a resource error.
         """
@@ -411,24 +525,28 @@ class XmlDocument(XMLResource):
             msg = "cannot create an ElementTree from a lazy resource"
             raise XMLResourceError(msg)
         elif hasattr(self._root, 'nsmap'):
-            return self._root.getroottree()
+            return self._root.getroottree()  # type: ignore[attr-defined]
         else:
             return ElementTree.ElementTree(self._root)
 
-    def tostring(self, indent='', max_lines=None, spaces_for_tab=4,
-                 xml_declaration=False, encoding='unicode', method='xml'):
+    def tostring(self, indent: str = '', max_lines: Optional[int] = None,
+                 spaces_for_tab: int = 4, xml_declaration: bool = False,
+                 encoding: str = 'unicode', method: str = 'xml') -> str:
         if self._lazy:
             raise XMLResourceError("cannot serialize a lazy XML document")
 
-        return etree_tostring(
+        _string = etree_tostring(
             elem=self._root,
             namespaces=self.namespaces,
             xml_declaration=xml_declaration,
             encoding=encoding,
             method=method
         )
+        if isinstance(_string, bytes):
+            return _string.decode('utf-8')
+        return _string
 
-    def decode(self, **kwargs):
+    def decode(self, **kwargs: Any) -> DecodeType[Any]:
         """
         Decode the XML document to a nested Python dictionary.
 
@@ -439,10 +557,16 @@ class XmlDocument(XMLResource):
         if 'namespaces' not in kwargs:
             kwargs['namespaces'] = self.namespaces
 
-        obj = (self.schema or self._fallback_schema).to_dict(self, **kwargs)
+        schema = self.schema or self._fallback_schema
+        if schema is None:
+            return None
+
+        obj = schema.to_dict(self, **kwargs)
         return obj[0] if isinstance(obj, tuple) else obj
 
-    def to_json(self, fp=None, json_options=None, **kwargs):
+    def to_json(self, fp: Optional[IO[str]] = None,
+                json_options: Optional[Dict[str, Any]] = None,
+                **kwargs: Any) -> JsonDecodeType:
         """
         Converts loaded XML data to a JSON string or file.
 
@@ -460,13 +584,18 @@ class XmlDocument(XMLResource):
         if 'decimal_type' not in kwargs:
             kwargs['decimal_type'] = float
 
-        errors = []
+        errors: List[XMLSchemaValidationError] = []
 
         if path is None and self._lazy and 'cls' not in json_options:
             json_options['cls'] = get_lazy_json_encoder(errors)
             kwargs['lazy_decode'] = True
 
-        obj = (self.schema or self._fallback_schema).decode(self, path=path, **kwargs)
+        schema = self.schema or self._fallback_schema
+        if schema is not None:
+            obj = schema.decode(self, path=path, **kwargs)
+        else:
+            obj = None
+
         if isinstance(obj, tuple):
             if fp is not None:
                 json.dump(obj[0], fp, **json_options)
@@ -484,13 +613,14 @@ class XmlDocument(XMLResource):
             result = json.dumps(obj, **json_options)
             return result if not errors else (result, tuple(errors))
 
-    def write(self, file, encoding='us-ascii', xml_declaration=None,
-              default_namespace=None, method="xml"):
+    def write(self, file: Union[str, TextIO, BinaryIO],
+              encoding: str = 'us-ascii', xml_declaration: bool = False,
+              default_namespace: Optional[str] = None, method: str = "xml") -> None:
         """Serialize an XML resource to a file. Cannot be used with lazy resources."""
         if self._lazy:
             raise XMLResourceError("cannot serialize a lazy XML document")
 
-        kwargs = {
+        kwargs: Dict[str, Any] = {
             'xml_declaration': xml_declaration,
             'encoding': encoding,
             'method': method,
@@ -498,19 +628,39 @@ class XmlDocument(XMLResource):
         if not default_namespace:
             kwargs['namespaces'] = self.namespaces
         else:
-            namespaces = self.namespaces.copy()
+            namespaces: Optional[Dict[Optional[str], str]]
+            if self.namespaces is None:
+                namespaces = {}
+            else:
+                namespaces = {k: v for k, v in self.namespaces.items()}
+
             if hasattr(self._root, 'nsmap'):
+                # noinspection PyTypeChecker
                 namespaces[None] = default_namespace
             else:
                 namespaces[''] = default_namespace
             kwargs['namespaces'] = namespaces
 
-        if hasattr(file, 'write'):
-            file.write(etree_tostring(self._root, **kwargs))
-            file.close()
-        elif encoding == 'unicode':
-            with open(file, 'w', encoding='utf-8') as fp:
-                fp.write(etree_tostring(self._root, **kwargs))
+        _string = etree_tostring(self._root, **kwargs)
+
+        if isinstance(file, str):
+            if isinstance(_string, str):
+                with open(file, 'w', encoding='utf-8') as fp:
+                    fp.write(_string)
+            else:
+                with open(file, 'wb') as _fp:
+                    _fp.write(_string)
+
+        elif isinstance(file, TextIOBase):
+            if isinstance(_string, bytes):
+                file.write(_string.decode('utf-8'))
+            else:
+                file.write(_string)
+
+        elif isinstance(file, IOBase):
+            if isinstance(_string, str):
+                file.write(_string.encode('utf-8'))
+            else:
+                file.write(_string)
         else:
-            with open(file, 'wb') as fp:
-                fp.write(etree_tostring(self._root, **kwargs))
+            raise XMLSchemaTypeError(f"unexpected type {type(file)} for 'file' argument")
