@@ -10,23 +10,38 @@
 """
 This module contains classes for XML Schema wildcards.
 """
+from typing import cast, Any, Callable, Dict, Iterable, Iterator, List, Optional, \
+    Tuple, Union, Counter
+
 from ..exceptions import XMLSchemaValueError
 from ..names import XSI_NAMESPACE, XSD_ANY, XSD_ANY_ATTRIBUTE, \
     XSD_OPEN_CONTENT, XSD_DEFAULT_OPEN_CONTENT, XSI_TYPE
-from ..helpers import get_namespace
-from ..xpath import XMLSchemaProxy, ElementPathMixin
+from ..aliases import ElementType, SchemaType, SchemaElementType, SchemaAttributeType, \
+    ModelGroupType, ModelParticleType, AtomicValueType, IterDecodeType, IterEncodeType, \
+    DecodedValueType, EncodedValueType
+from ..helpers import get_namespace, raw_xml_encode
+from ..xpath import XMLSchemaProtocol, ElementProtocol, XMLSchemaProxy, ElementPathMixin
 from .xsdbase import ValidationMixin, XsdComponent
 from .particles import ParticleMixin
+from . import elements
 
 
-class XsdWildcard(XsdComponent, ValidationMixin):
+OccursCounterType = Counter[Union[ModelParticleType, Tuple[ModelParticleType]]]
+
+
+class XsdWildcard(XsdComponent):
     names = ()
-    namespace = ('##any',)
-    not_namespace = ()
-    not_qname = ()
+    namespace: Union[Tuple[str], List[str]] = ('##any',)
+    not_namespace: Union[Tuple[()], List[str]] = ()
+    not_qname: Union[Tuple[()], List[str]] = ()
     process_contents = 'strict'
 
-    def __repr__(self):
+    # For compatibility with protocol of XSD elements/attributes
+    type = None
+    default = None
+    fixed = None
+
+    def __repr__(self) -> str:
         if self.not_namespace:
             return '%s(not_namespace=%r, process_contents=%r)' % (
                 self.__class__.__name__, self.not_namespace, self.process_contents
@@ -36,11 +51,9 @@ class XsdWildcard(XsdComponent, ValidationMixin):
                 self.__class__.__name__, self.namespace, self.process_contents
             )
 
-    def _parse(self):
-        super(XsdWildcard, self)._parse()
-
+    def _parse(self) -> None:
         # Parse namespace and processContents
-        namespace = self.elem.get('namespace', '##any').strip()
+        namespace = self.elem.attrib.get('namespace', '##any').strip()
         if namespace == '##any':
             pass
         elif not namespace:
@@ -63,7 +76,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
                 else:
                     self.namespace.append(ns)
 
-        process_contents = self.elem.get('processContents', 'strict')
+        process_contents = self.elem.attrib.get('processContents', 'strict')
         if process_contents == 'strict':
             pass
         elif process_contents not in ('lax', 'skip'):
@@ -72,7 +85,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         else:
             self.process_contents = process_contents
 
-    def _parse_not_constraints(self):
+    def _parse_not_constraints(self) -> None:
         if 'notNamespace' not in self.elem.attrib:
             pass
         elif 'namespace' in self.elem.attrib:
@@ -127,10 +140,12 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         self.not_qname = names
 
     @property
-    def built(self):
+    def built(self) -> bool:
         return True
 
-    def is_matching(self, name, default_namespace=None, **kwargs):
+    def is_matching(self, name: Optional[str],
+                    default_namespace: Optional[str] = None,
+                    **kwargs: Any) -> bool:
         if name is None:
             return False
         elif not name or name[0] == '{':
@@ -141,17 +156,19 @@ class XsdWildcard(XsdComponent, ValidationMixin):
             return self.is_namespace_allowed('') or \
                 self.is_namespace_allowed(default_namespace)
 
-    def is_namespace_allowed(self, namespace):
+    def is_namespace_allowed(self, namespace: str) -> bool:
         if self.not_namespace:
             return namespace not in self.not_namespace
         elif '##any' in self.namespace or namespace == XSI_NAMESPACE:
             return True
         elif '##other' in self.namespace:
-            return namespace and namespace != self.target_namespace
+            if not namespace:
+                return False
+            return namespace != self.target_namespace
         else:
             return namespace in self.namespace
 
-    def deny_namespaces(self, namespaces):
+    def deny_namespaces(self, namespaces: List[str]) -> bool:
         if self.not_namespace:
             return all(x in self.not_namespace for x in namespaces)
         elif '##any' in self.namespace:
@@ -161,7 +178,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         else:
             return all(x not in self.namespace for x in namespaces)
 
-    def deny_qnames(self, names):
+    def deny_qnames(self, names: Iterable[str]) -> bool:
         if self.not_namespace:
             return all(x in self.not_qname or get_namespace(x) in self.not_namespace
                        for x in names)
@@ -174,13 +191,17 @@ class XsdWildcard(XsdComponent, ValidationMixin):
             return all(x in self.not_qname or get_namespace(x) not in self.namespace
                        for x in names)
 
-    def is_restriction(self, other, check_occurs=True):
+    def is_restriction(self, other: Union[ModelParticleType, 'XsdAnyAttribute'],
+                       check_occurs: bool = True) -> bool:
         if check_occurs and isinstance(self, ParticleMixin) \
+                and not isinstance(other, XsdAnyAttribute) \
                 and not self.has_occurs_restriction(other):
             return False
-        elif not isinstance(other, type(self)):
+        elif not isinstance(other, self.__class__):
             return False
-        elif other.process_contents == 'strict' and self.process_contents != 'strict':
+
+        other: XsdWildcard  # type: ignore[no-redef]
+        if other.process_contents == 'strict' and self.process_contents != 'strict':
             return False
         elif other.process_contents == 'lax' and self.process_contents == 'skip':
             return False
@@ -226,7 +247,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         else:
             return all(ns in other.namespace for ns in self.namespace)
 
-    def union(self, other):
+    def union(self, other: Union['XsdAnyElement', 'XsdAnyAttribute']) -> None:
         """
         Update an XSD wildcard with the union of itself and another XSD wildcard.
         """
@@ -267,6 +288,8 @@ class XsdWildcard(XsdComponent, ValidationMixin):
             self.namespace = ['##any'] if not self.not_namespace else []
             return
 
+        w1: XsdWildcard
+        w2: XsdWildcard
         if '##any' in self.namespace or self.namespace == other.namespace:
             return
         elif '##any' in other.namespace:
@@ -277,6 +300,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         elif '##other' in self.namespace:
             w1, w2 = self, other
         else:
+            assert isinstance(self.namespace, list)
             self.namespace.extend(ns for ns in other.namespace if ns not in self.namespace)
             return
 
@@ -291,7 +315,7 @@ class XsdWildcard(XsdComponent, ValidationMixin):
             self.namespace = []
             self.not_namespace = ['', w1.target_namespace] if w1.target_namespace else ['']
 
-    def intersection(self, other):
+    def intersection(self, other: Union['XsdAnyElement', 'XsdAnyAttribute']) -> None:
         """
         Update an XSD wildcard with the intersection of itself and another XSD wildcard.
         """
@@ -342,19 +366,16 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         elif '##other' not in other.namespace:
             self.namespace = [ns for ns in self.namespace if ns in other.namespace]
         else:
+            assert isinstance(self.namespace, list)
             if other.target_namespace in self.namespace:
                 self.namespace.remove(other.target_namespace)
             if '' in self.namespace:
                 self.namespace.remove('')
 
-    def iter_decode(self, source, validation='lax', **kwargs):
-        raise NotImplementedError
 
-    def iter_encode(self, obj, validation='lax', **kwargs):
-        raise NotImplementedError
-
-
-class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
+class XsdAnyElement(XsdWildcard, ParticleMixin,
+                    ElementPathMixin[SchemaElementType],
+                    ValidationMixin[ElementType, Any]):
     """
     Class for XSD 1.0 *any* wildcards.
 
@@ -369,31 +390,38 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
         </any>
     """
     _ADMITTED_TAGS = {XSD_ANY}
-    precedences = ()
+    precedences: Dict[ModelGroupType, List[ModelParticleType]]
+    copy: Callable[['XsdAnyElement'], 'XsdAnyElement']
 
-    def __init__(self, elem, schema, parent, maps=None):
-        super(XsdAnyElement, self).__init__(elem, schema, parent, maps)
-        ElementPathMixin.__init__(self)
+    def __init__(self, elem: ElementType, schema: SchemaType, parent: XsdComponent) -> None:
+        self.precedences = {}
+        super(XsdAnyElement, self).__init__(elem, schema, parent)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         if self.namespace:
             return '%s(namespace=%r, process_contents=%r, occurs=%r)' % (
-                self.__class__.__name__, self.namespace, self.process_contents, self.occurs
+                self.__class__.__name__, self.namespace,
+                self.process_contents, list(self.occurs)
             )
         else:
             return '%s(not_namespace=%r, process_contents=%r, occurs=%r)' % (
-                self.__class__.__name__, self.not_namespace, self.process_contents, self.occurs
+                self.__class__.__name__, self.not_namespace,
+                self.process_contents, list(self.occurs)
             )
 
     @property
-    def xpath_proxy(self):
-        return XMLSchemaProxy(self.schema, self)
+    def xpath_proxy(self) -> XMLSchemaProxy:
+        return XMLSchemaProxy(
+            schema=cast(XMLSchemaProtocol, self.schema),
+            base_element=cast(ElementProtocol, self)
+        )
 
-    def _parse(self):
+    def _parse(self) -> None:
         super(XsdAnyElement, self)._parse()
         self._parse_particle(self.elem)
 
-    def match(self, name, default_namespace=None, resolve=False, **kwargs):
+    def match(self, name: Optional[str], default_namespace: Optional[str] = None,
+              resolve: bool = False, **kwargs: Any) -> Optional[SchemaElementType]:
         """
         Returns the element wildcard if name is matching the name provided
         as argument, `None` otherwise.
@@ -405,8 +433,8 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
         resolve and return the element matching the name.
         :param kwargs: additional options used by XSD 1.1 xs:any wildcards.
         """
-        if not self.is_matching(name, default_namespace, **kwargs):
-            return
+        if not name or not self.is_matching(name, default_namespace, **kwargs):
+            return None
         elif not resolve:
             return self
 
@@ -416,55 +444,63 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
             else:
                 return self.maps.lookup_element(name)
         except LookupError:
-            pass
+            return None
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Any]:
         return iter(())
 
-    def iter(self, tag=None):
+    def iter(self, tag: Optional[str] = None) -> Iterator[Any]:
         return iter(())
 
-    def iterchildren(self, tag=None):
+    def iterchildren(self, tag: Optional[str] = None) -> Iterator[Any]:
         return iter(())
 
     @staticmethod
-    def iter_substitutes():
+    def iter_substitutes() -> Iterator[Any]:
         return iter(())
 
-    def iter_decode(self, elem, validation='lax', **kwargs):
-        if not self.is_matching(elem.tag):
-            reason = "{!r} is not allowed here".format(elem)
-            yield self.validation_error(validation, reason, elem, **kwargs)
+    def iter_decode(self, obj: ElementType, validation: str = 'lax', **kwargs: Any) \
+            -> IterDecodeType[Any]:
 
-        elif self.process_contents == 'skip':
+        if not self.is_matching(obj.tag):
+            reason = "{!r} is not allowed here".format(obj)
+            yield self.validation_error(validation, reason, obj, **kwargs)
+
+        if self.process_contents == 'skip':
+            if 'process_skipped' not in kwargs or not kwargs['process_skipped']:
+                return
+
+        namespace = get_namespace(obj.tag)
+        if not self.maps.load_namespace(namespace):
+            reason = f"unavailable namespace {namespace!r}"
+        else:
+            try:
+                xsd_element = self.maps.lookup_element(obj.tag)
+            except LookupError:
+                reason = f"element {obj.tag!r} not found"
+            else:
+                yield from xsd_element.iter_decode(obj, validation, **kwargs)
+                return
+
+        if XSI_TYPE in obj.attrib:
+            if self.process_contents == 'strict':
+                xsd_element = self.maps.validator.create_element(
+                    obj.tag, parent=self, form='unqualified'
+                )
+            else:
+                xsd_element = self.maps.validator.create_element(
+                    obj.tag, parent=self, nillable='true', form='unqualified'
+                )
+
+            yield from xsd_element.iter_decode(obj, validation, **kwargs)
             return
 
-        elif self.maps.load_namespace(get_namespace(elem.tag)):
-            try:
-                xsd_element = self.maps.lookup_element(elem.tag)
-            except LookupError:
-                if XSI_TYPE in elem.attrib:
-                    if self.process_contents == 'lax':
-                        xsd_element = self.maps.validator.create_element(elem.tag, nillable='true')
-                    else:
-                        xsd_element = self.maps.validator.create_element(elem.tag)
-                    yield from xsd_element.iter_decode(elem, validation, **kwargs)
-                elif validation == 'skip' or self.process_contents == 'lax':
-                    yield from self.any_type.iter_decode(elem, validation, **kwargs)
-                else:
-                    reason = "element %r not found." % elem.tag
-                    yield self.validation_error(validation, reason, elem, **kwargs)
-            else:
-                yield from xsd_element.iter_decode(elem, validation, **kwargs)
+        if validation != 'skip' and self.process_contents == 'strict':
+            yield self.validation_error(validation, reason, obj, **kwargs)
+        yield from self.any_type.iter_decode(obj, validation, **kwargs)
 
-        elif validation == 'skip':
-            yield self.any_type.decode(elem) if len(elem) > 0 else elem.text
-
-        elif self.process_contents == 'strict':
-            reason = "unavailable namespace {!r}".format(get_namespace(elem.tag))
-            yield self.validation_error(validation, reason, elem, **kwargs)
-
-    def iter_encode(self, obj, validation='lax', **kwargs):
+    def iter_encode(self, obj: Tuple[str, ElementType], validation: str = 'lax', **kwargs: Any) \
+            -> IterEncodeType[Any]:
         name, value = obj
         namespace = get_namespace(name)
 
@@ -472,32 +508,59 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
             reason = "element {!r} is not allowed here".format(name)
             yield self.validation_error(validation, reason, value, **kwargs)
 
-        elif self.process_contents == 'skip':
-            return
+        if self.process_contents == 'skip':
+            if 'process_skipped' not in kwargs or not kwargs['process_skipped']:
+                return
 
-        elif self.maps.load_namespace(namespace):
+        if not self.maps.load_namespace(namespace):
+            reason = f"unavailable namespace {namespace!r}"
+        else:
             try:
                 xsd_element = self.maps.lookup_element(name)
             except LookupError:
-                if validation == 'skip' or self.process_contents == 'lax':
-                    yield from self.any_type.iter_encode(obj, validation, **kwargs)
-                elif self.process_contents == 'strict':
-                    reason = "element %r not found." % name
-                    yield self.validation_error(validation, reason, **kwargs)
+                reason = f"element {name!r} not found"
             else:
                 yield from xsd_element.iter_encode(value, validation, **kwargs)
+                return
 
-        elif validation == 'skip':
-            yield self.any_type.encode(value)
+        # Check if there is an xsi:type attribute, but it has to extract
+        # attributes using the converter instance.
+        if self.process_contents == 'strict':
+            xsd_element = self.maps.validator.create_element(
+                name, parent=self, form='unqualified'
+            )
+        else:
+            xsd_element = self.maps.validator.create_element(
+                name, parent=self, nillable='true', form='unqualified'
+            )
 
-        elif self.process_contents == 'strict':
-            reason = "unavailable namespace {!r}".format(namespace)
+        try:
+            converter = kwargs['converter']
+        except KeyError:
+            converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
+
+        try:
+            level = kwargs['level']
+        except KeyError:
+            element_data = converter.element_encode(value, xsd_element)
+        else:
+            element_data = converter.element_encode(value, xsd_element, level)
+
+        if XSI_TYPE in element_data.attributes:
+            yield from xsd_element.iter_encode(value, validation, **kwargs)
+            return
+
+        if validation != 'skip' and self.process_contents == 'strict':
             yield self.validation_error(validation, reason, **kwargs)
+        yield from self.any_type.iter_encode(obj, validation, **kwargs)
 
-    def is_overlap(self, other):
+    def is_overlap(self, other: ModelParticleType) -> bool:
         if not isinstance(other, XsdAnyElement):
-            return other.is_overlap(self)
-        elif self.not_namespace:
+            if isinstance(other, elements.XsdElement):
+                return other.is_overlap(self)
+            return False
+
+        if self.not_namespace:
             if other.not_namespace:
                 return True
             elif '##any' in other.namespace:
@@ -524,11 +587,11 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
         else:
             return any(ns in self.namespace for ns in other.namespace)
 
-    def is_consistent(self, other):
+    def is_consistent(self, other: SchemaElementType, **kwargs: Any) -> bool:
         return True
 
 
-class XsdAnyAttribute(XsdWildcard):
+class XsdAnyAttribute(XsdWildcard, ValidationMixin[Tuple[str, str], DecodedValueType]):
     """
     Class for XSD 1.0 *anyAttribute* wildcards.
 
@@ -540,9 +603,15 @@ class XsdAnyAttribute(XsdWildcard):
           Content: (annotation?)
         </anyAttribute>
     """
+    copy: Callable[['XsdAnyAttribute'], 'XsdAnyAttribute']
     _ADMITTED_TAGS = {XSD_ANY_ATTRIBUTE}
 
-    def match(self, name, default_namespace=None, resolve=False, **kwargs):
+    # Added for compatibility with protocol of XSD attributes
+    use = None
+    inheritable = False  # XSD 1.1 attributes
+
+    def match(self, name: Optional[str], default_namespace: Optional[str] = None,
+              resolve: bool = False, **kwargs: Any) -> Optional[SchemaAttributeType]:
         """
         Returns the attribute wildcard if name is matching the name provided
         as argument, `None` otherwise.
@@ -554,8 +623,8 @@ class XsdAnyAttribute(XsdWildcard):
         resolve and return the attribute matching the name.
         :param kwargs: additional options that can be used by certain components.
         """
-        if not self.is_matching(name, default_namespace, **kwargs):
-            return
+        if not name or not self.is_matching(name, default_namespace, **kwargs):
+            return None
         elif not resolve:
             return self
 
@@ -565,67 +634,67 @@ class XsdAnyAttribute(XsdWildcard):
             else:
                 return self.maps.lookup_attribute(name)
         except LookupError:
-            pass
+            return None
 
-    def iter_decode(self, attribute, validation='lax', **kwargs):
-        name, value = attribute
+    def iter_decode(self, obj: Tuple[str, str], validation: str = 'lax', **kwargs: Any) \
+            -> IterDecodeType[DecodedValueType]:
+        name, value = obj
 
         if not self.is_matching(name):
             reason = "attribute %r not allowed." % name
-            yield self.validation_error(validation, reason, attribute, **kwargs)
+            yield self.validation_error(validation, reason, obj, **kwargs)
 
-        elif self.process_contents == 'skip':
-            return
+        if self.process_contents == 'skip':
+            if 'process_skipped' not in kwargs or not kwargs['process_skipped']:
+                return
 
-        elif self.maps.load_namespace(get_namespace(name)):
+        if self.maps.load_namespace(get_namespace(name)):
             try:
                 xsd_attribute = self.maps.lookup_attribute(name)
             except LookupError:
-                if validation == 'skip':
-                    yield value
-                elif self.process_contents == 'strict':
+                if validation != 'skip' and self.process_contents == 'strict':
                     reason = "attribute %r not found." % name
-                    yield self.validation_error(validation, reason, attribute, **kwargs)
+                    yield self.validation_error(validation, reason, obj, **kwargs)
             else:
                 yield from xsd_attribute.iter_decode(value, validation, **kwargs)
+                return
 
-        elif validation == 'skip':
-            yield value
-
-        elif self.process_contents == 'strict':
+        elif validation != 'skip' and self.process_contents == 'strict':
             reason = "unavailable namespace {!r}".format(get_namespace(name))
             yield self.validation_error(validation, reason, **kwargs)
 
-    def iter_encode(self, attribute, validation='lax', **kwargs):
-        name, value = attribute
+        yield value
+
+    def iter_encode(self, obj: Tuple[str, AtomicValueType], validation: str = 'lax',
+                    **kwargs: Any) -> IterEncodeType[EncodedValueType]:
+        name, value = obj
         namespace = get_namespace(name)
 
         if not self.is_namespace_allowed(namespace):
             reason = "attribute %r not allowed." % name
-            yield self.validation_error(validation, reason, attribute, **kwargs)
+            yield self.validation_error(validation, reason, obj, **kwargs)
 
-        elif self.process_contents == 'skip':
-            return
+        if self.process_contents == 'skip':
+            if 'process_skipped' not in kwargs or not kwargs['process_skipped']:
+                return
 
-        elif self.maps.load_namespace(namespace):
+        if self.maps.load_namespace(namespace):
             try:
                 xsd_attribute = self.maps.lookup_attribute(name)
             except LookupError:
-                if validation == 'skip':
-                    yield str(value)
-                elif self.process_contents == 'strict':
+                if validation != 'skip' and self.process_contents == 'strict':
                     reason = "attribute %r not found." % name
-                    yield self.validation_error(validation, reason, attribute, **kwargs)
+                    yield self.validation_error(validation, reason, obj, **kwargs)
             else:
                 yield from xsd_attribute.iter_encode(value, validation, **kwargs)
+                return
 
-        elif validation == 'skip':
-            yield str(value)
-
-        elif self.process_contents == 'strict':
+        elif validation != 'skip' and self.process_contents == 'strict':
             reason = "unavailable namespace {!r}".format(get_namespace(name))
             yield self.validation_error(validation, reason, **kwargs)
 
+        yield raw_xml_encode(value)
+
 
 class Xsd11AnyElement(XsdAnyElement):
     """
@@ -643,11 +712,15 @@ class Xsd11AnyElement(XsdAnyElement):
           Content: (annotation?)
         </any>
     """
-    def _parse(self):
+    def _parse(self) -> None:
         super(Xsd11AnyElement, self)._parse()
         self._parse_not_constraints()
 
-    def is_matching(self, name, default_namespace=None, group=None, occurs=None):
+    def is_matching(self, name: Optional[str],
+                    default_namespace: Optional[str] = None,
+                    group: Optional[ModelGroupType] = None,
+                    occurs: Optional[OccursCounterType] = None,
+                    **kwargs: Any) -> bool:
         """
         Returns `True` if the component name is matching the name provided as argument,
         `False` otherwise. For XSD elements the matching is extended to substitutes.
@@ -690,15 +763,13 @@ class Xsd11AnyElement(XsdAnyElement):
 
         return name not in self.not_qname
 
-    def is_consistent(self, other):
+    def is_consistent(self, other: SchemaElementType, **kwargs: Any) -> bool:
         if isinstance(other, XsdAnyElement) or self.process_contents == 'skip':
             return True
         xsd_element = self.match(other.name, other.default_namespace, resolve=True)
         return xsd_element is None or other.is_consistent(xsd_element, strict=False)
 
-    def add_precedence(self, other, group):
-        if not self.precedences:
-            self.precedences = {}
+    def add_precedence(self, other: ModelParticleType, group: ModelGroupType) -> None:
         try:
             self.precedences[group].append(other)
         except KeyError:
@@ -719,13 +790,13 @@ class Xsd11AnyAttribute(XsdAnyAttribute):
           Content: (annotation?)
         </anyAttribute>
     """
-    inheritable = False  # Added for reduce checkings on XSD 1.1 attributes
-
-    def _parse(self):
+    def _parse(self) -> None:
         super(Xsd11AnyAttribute, self)._parse()
         self._parse_not_constraints()
 
-    def is_matching(self, name, default_namespace=None, **kwargs):
+    def is_matching(self, name: Optional[str],
+                    default_namespace: Optional[str] = None,
+                    **kwargs: Any) -> bool:
         if name is None:
             return False
         elif not name or name[0] == '{':
@@ -737,8 +808,13 @@ class Xsd11AnyAttribute(XsdAnyAttribute):
             namespace = default_namespace
 
         if '##defined' in self.not_qname and name in self.maps.attributes:
-            if self.maps.attributes[name].schema is self.schema:
+            xsd_attribute = self.maps.attributes[name]
+            if isinstance(xsd_attribute, tuple):
+                if xsd_attribute[1] is self.schema:
+                    return False
+            elif xsd_attribute.schema is self.schema:
                 return False
+
         return name not in self.not_qname and self.is_namespace_allowed(namespace)
 
 
@@ -755,15 +831,15 @@ class XsdOpenContent(XsdComponent):
     """
     _ADMITTED_TAGS = {XSD_OPEN_CONTENT}
     mode = 'interleave'
-    any_element = None
+    any_element = None  # type: Xsd11AnyElement
 
-    def __init__(self, elem, schema, parent):
+    def __init__(self, elem: ElementType, schema: SchemaType, parent: XsdComponent) -> None:
         super(XsdOpenContent, self).__init__(elem, schema, parent)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return '%s(mode=%r)' % (self.__class__.__name__, self.mode)
 
-    def _parse(self):
+    def _parse(self) -> None:
         super(XsdOpenContent, self)._parse()
         try:
             self.mode = self.elem.attrib['mode']
@@ -787,10 +863,10 @@ class XsdOpenContent(XsdComponent):
             self.any_element = any_element
 
     @property
-    def built(self):
+    def built(self) -> bool:
         return True
 
-    def is_restriction(self, other):
+    def is_restriction(self, other: 'XsdOpenContent') -> bool:
         if other is None or other.mode == 'none':
             return self.mode == 'none'
         elif self.mode == 'interleave' and other.mode == 'suffix':
@@ -814,10 +890,10 @@ class XsdDefaultOpenContent(XsdOpenContent):
     _ADMITTED_TAGS = {XSD_DEFAULT_OPEN_CONTENT}
     applies_to_empty = False
 
-    def __init__(self, elem, schema):
+    def __init__(self, elem: ElementType, schema: SchemaType) -> None:
         super(XsdOpenContent, self).__init__(elem, schema)
 
-    def _parse(self):
+    def _parse(self) -> None:
         super(XsdDefaultOpenContent, self)._parse()
         if self.parent is not None:
             self.parse_error("defaultOpenContent must be a child of the schema")
